-- ==== HARDWARE LED CONTROL === --
LEDCtl = {
	DEVICE_ADDRESS = 0x40,
	PRESCALE = 0x1e,
	GAMMA = 0.45455,
	-- phase shifting gets flickering:
	-- https://community.nxp.com/message/919888
	PHASE = 0,

	register = {
		MODE1 = 0x00,
		MODE2 = 0x01,
		PRE_SCALE = 0xfe,
		ALL_LED_OFF_L = 0xfc,
		ALL_LED_OFF_H = 0xfd,
		LED_ON_L = 0x06,
		LED_ON_H = 0x07,
		LED_OFF_L = 0x08,
		LED_OFF_H = 0x09
	},

	i2c0 = {
		ID = 0,
		-- SDA on old boards connected to GPIO16
		-- but can't be used for i2c. Need bodge-wire
		-- to GPIO2
		SDA = 4,	-- GPIO2
		SCL = 5,	-- GPIO14
		SPEED = i2c.FAST
	},

	dimmer_table = {}
}

function LEDCtl:init()
	gpio.mode( 0, gpio.INPUT, gpio.FLOAT )
	self.i2c0.SPEED = i2c.setup( 0, self.i2c0.SDA, self.i2c0.SCL, self.i2c0.SPEED )

	-- Set PRESCALE to 200 Hz 25 MHz / ( 4096 * 200 ) - 1 = 30
	LEDCtl:write( self.register.PRE_SCALE, self.PRESCALE )

	-- set MODE0 to 0:
	--  * does not respond to LED All Call bus address
	--  * does not respond to subaddresses
	--  * SLEEP: normal mode
	--  * Auto increment disabled
	--  * Internal clock
	LEDCtl:write( self.register.MODE1, 0 )

	-- OUTDRV 1, directly drive mosfet
	LEDCtl:write( self.register.MODE2, 0x04 )

	-- Update dimmer_table with pre-calculated values
	LEDCtl:set_gamma( self.GAMMA )

	-- We can stagger the ON counter so LEDs don't get turned on all at once
	-- however there is a bug in the silicon of the PCA9685, if LED_ON > LED_OFF
	-- the PWM output misses a whole cycle after setting the LED_OFF register.
	-- See comment at PHASE variable.
	for led = 0, 1, 15 do
		-- Send first 8 bits, then send last 8 bits of led_on
		LEDCtl:write( self.register.LED_ON_L + 4 * led, bit.band( led * self.PHASE, 0xff ) )
		LEDCtl:write( self.register.LED_ON_H + 4 * led, bit.rshift( led * self.PHASE, 8 ) )
	end

	return true
end

function LEDCtl:write( reg, data )
	i2c.start( self.i2c0.ID )
	i2c.address( self.i2c0.ID, self.DEVICE_ADDRESS, i2c.TRANSMITTER )
	i2c.write( self.i2c0.ID, reg )
	ret = i2c.write( self.i2c0.ID, data )
	i2c.stop( self.i2c0.ID )
	return ret
end

function LEDCtl:read( reg, length )
	i2c.start( self.i2c0.ID )
	i2c.address( self.i2c0.ID, self.DEVICE_ADDRESS, i2c.TRANSMITTER )
	i2c.write( self.i2c0.ID, reg )
	i2c.stop( self.i2c0.ID )
	i2c.start( self.i2c0.ID )
	i2c.address( self.i2c0.ID, self.DEVICE_ADDRESS, i2c.RECEIVER )
	ret = i2c.read( self.i2c0.ID, length )
	i2c.stop( self.i2c0.ID )
	return ret
end

function LEDCtl:set_gamma( gamma )
	self.GAMMA = gamma
	-- https://www.mikrocontroller.net/articles/Diskussion:LED-Fading#Diskussion_wissenschaftl.-technischer_Hintergrund
	for i = 1, 100, 1 do
		self.dimmer_table[i] = math.floor( 4095 * ( i / 100 )^( 1 / self.GAMMA ) + 0.5 )
		--print( i..";"..self.dimmer_table[i] )
	end
	return true
end

function LEDCtl:set_led( led, duty )
	if( led < 0 or led > 15 ) then
		return false
	end
	if( duty < 0 or duty > 100 ) then
		return false
	end

	if( duty == 0 ) then
		-- Set LED_OFF register same as LED_ON
		LEDCtl:write( self.register.LED_OFF_L + 4 * led, bit.band( self.PHASE * led, 0xff ) )
		LEDCtl:write( self.register.LED_OFF_H + 4 * led, bit.rshift( self.PHASE * led, 8 ) )
	else
		local led_on = self.PHASE * led
		local led_off =  self.dimmer_table[ duty ] + led_on
		if( led_off > 4095 ) then
			led_off = led_off - 4096
		end

		-- Write LED_OFF register
		LEDCtl:write( self.register.LED_OFF_L + 4 * led, bit.band( led_off, 0xff ) )
		LEDCtl:write( self.register.LED_OFF_H + 4 * led, bit.rshift( led_off, 8 ) )
	end
	return true
end

-- re-implemented in a functional way.
-- doesn't run stable with OOP :(

led_cur = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
led_start = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
led_target = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
speed = 40
max_steps = 0
step = 0
runs = false

fade_timer = tmr.create()


function led_register( led, new_brightness )
	led_start[ led ] = led_cur[ led ]
	led_target[ led ] = new_brightness
end

function next_step( led, step, max )
	if( led_cur[ led ] == led_target[ led ] ) then
		return false
	end

		local new_brightness = math.floor( (( led_target[ led ] - led_start[ led ] ) / max) * step + led_start[ led ] )
	--print( "(("..self.target.." - "..self.start.." ) / "..max..") * "..step.." + "..self.start.." = "..new_brightness )

	if( new_brightness < 0 ) then
		new_brightness = led_target[ led ]
	elseif( ( led_target[ led ] - led_start[ led ] > 0 and new_brightness > led_target[ led ] ) or ( led_target[ led ] - led_start[ led ] < 0 and new_brightness < led_target[ led ] ) ) then
		new_brightness = led_target[ led ]
	end

	LEDCtl:set_led( led - 1, new_brightness )
	return true
end

function setSpeed( s )
	if( s < 5 or s > 10000 ) then
		return false
	end
	speed = s
	return true
end

function fade_start()
	-- can't run a fader if it's already running
	if( runs ) then
		return false
	end
	runs = true

	-- start by retrieving the max number of steps for the job
	local max = 0
	for i = 1, 16, 1 do
		if( fadeMe( i ) ) then
			if( math.abs( led_start[ i ] - led_target[ i ] ) > max ) then
				max = math.abs( led_start[ i ] - led_target[ i ] )
			end
		end
	end
	max_steps = max
	step = 0

	fade_timer:register( speed, tmr.ALARM_AUTO, function( t )
		for i = 1, 16, 1 do
			if( fadeMe( i ) ) then
				next_step( i, step, max_steps )
			end
		end
		step = step + 1

		if( step > max_steps ) then
			t:stop()

			runs = false
			-- save target value as new start value.
			for i = 1, 16, 1 do
				led_cur[ i ] = led_target[ i ]
			end
		end
	end)
	fade_timer:start()
end

function fadeMe( led )
	return( led_cur[ led ] ~= led_target[ led ] )
end

function fade_stop()
	fade_timer:unregister()

	-- Update start values with current values, so we can resume dimming
	for i = 1, 16, 1 do
		led_cur[ i ] = led_target[ i ]
	end

	runs = false
end

function running()
	return runs
end

function setInit( led, brightness )
	led_cur[ led ] = brightness
	return true
end
